#include "rtk_measure.h"

namespace Localization {
namespace Ego {

double RtkMeasure::latestUpdateTime_ = -1.0;

RtkMeasure::RtkMeasure(const MeasurementType &mType, const InsInfo &insInfo)
:BaseMeasure(mType)
{
  isRtkBlock_ = false;
  isKFUsed_ = false;
  stamp_ = insInfo.stamp_;
  seq_ = insInfo.seq_;
  type_ = mType;
  positionType_ = insInfo.positionType_;
  solutionStatus = insInfo.solutionStatus;
  satUseNum_ = insInfo.satUseNum_;
  // pos
  posLLA_.x() = insInfo.longitude_;
  posLLA_.y() = insInfo.latitude_;
  posLLA_.z() = insInfo.elevation_;
  Eigen::Vector3d mapOffset_{0,0,0};
  posUTM_ = insInfo.utmPosition_ - mapOffset_;
  posUTM_.z() = 0.0;
  attitude_ = insInfo.attitude_;
  Eigen::Vector3d ypr(attitude_.z(), 0.0, 0.0);
  quatUTM_ = EulerYPR2RotYPR(ypr);
  {
    rtkPos_ = posUTM_;
    rtkAtt_ = attitude_;
  }
  velocityVeh_ = insInfo.velocityVeh_;
  veloENU_ = insInfo.velocityENU_;
  heading_ = insInfo.heading;
  cog_ = insInfo.cog;

  mRm_ = Eigen::MatrixXd::Zero(6, 6);
  mRm_.block<3, 3>(0, 0) = gnssRVar_ * Eigen::Matrix3d::Identity();
  mRm_.block<3, 3>(3, 3) = gnssPVar_ * Eigen::Matrix3d::Identity();
  CheckValid();
};

void RtkMeasure::CheckValid()
{
  isRtkValid_ = true;
  interval_ = stamp_ - latestUpdateTime_;
  if (interval_ < 1e-5) {
    std::cerr<<std::setprecision(15)<<"RtkMeasure():time error,this frame is repeated,curstamp"<<stamp_
             <<",latestUpdateTime_"<<latestUpdateTime_<<std::endl;
  } else if (interval_ < 1.5) {
    // normal
  } else if (interval_ < 1e2) {
    isRtkValid_ = false;
    std::cerr<<std::setprecision(15)<<"RtkMeasure():time error, rtk time jumps, interval_ is:"<<interval_<<
      ",set isRtkValid_ false,curstamp"<<stamp_<<",latestUpdateTime_"<<latestUpdateTime_<<std::endl;
  } else {
    interval_ = 0.0;
    isRtkValid_ = false;
    std::cerr<<std::setprecision(15)<<"RtkMeasure():time error,time interval_ is bigger than 100s,so set interval to 0.0,"
        "set isRtkValid_ false,curstamp"<<stamp_<<",latestUpdateTime_"<<latestUpdateTime_<<std::endl;
  }
  latestUpdateTime_ = stamp_;
  if (fabs(lastRtkStamp_ - lastImuStamp_) > 2.0) {
    std::cerr<<std::setprecision(15)<<"RtkMeasure():time error,rtk stamp has big diff with imu stamp,lastRtkStamp:"<<
      lastRtkStamp_<<",lastImuStamp:"<<lastImuStamp_<<",curRtkStamp:"<<stamp_<<std::endl;
    isRtkBlock_ = true;
  } else {
    isRtkBlock_ = false;
  }
  lastRtkStamp_ = stamp_;

  if (positionType_ == 4U) {
    if (solutionStatus == 3U) {
      if (satUseNum_ >= 10) {
        mRm_.block<3, 3>(0, 0) = 1 * gnssRVar_ * Eigen::Matrix3d::Identity();
        mRm_.block<3, 3>(3, 3) = 1 * gnssPVar_ * Eigen::Matrix3d::Identity();
      }
    } else {
      std::cerr<<std::setprecision(15)<<"RtkMeasure-CheckValid():status error,Rtk的positionType_为4,但是solutionStatus is:"<<solutionStatus<<
        ",set isRtkValid_ false,观测方程不使用RTK,时间戳:"<<stamp_<<std::endl;
      isRtkValid_ = false;
    }
  } else {
    std::cerr<<std::setprecision(15)<<"RtkMeasure-CheckValid():status error,Rtk的positionType_不为4,positionType_ is:"<<positionType_<<
      ",set isRtkValid_ false,观测方程不使用RTK,时间戳:"<<stamp_<<std::endl;
    isRtkValid_ = false;
  }
}

bool RtkMeasure::Update(const State &state, State &nextState)
{
  nextState = state;
  if (isRtkValid_ == false) {
    std::cerr<<std::setprecision(15)<<"rtk-Update():error,isRtkValid_ false,时间戳:"<<stamp_<<
      "can't update,use next pridict"<<std::endl;
    return false;
  }
  nextState.time = stamp_;

  Eigen::VectorXd innov(6);
  Eigen::MatrixXd mH = Eigen::MatrixXd::Zero(6, StateIndex::STATE_TOTAL);
  ObserveRTK(state,nextState,innov,mH);
  if (!UpdateStateGnss(state, nextState, innov, mH)) {
    return false;
  }

  return true;
}

bool RtkMeasure::UpdateStateGnss(const State &state, State &nextState, const Eigen::VectorXd &innov,
     const Eigen::MatrixXd &mH)
{
  Eigen::MatrixXd mS = mH * state.mP * mH.transpose() + mRm_;
  Eigen::MatrixXd mSInv = mS.inverse();
  Eigen::MatrixXd mK = state.mP * mH.transpose() * mSInv;
  Eigen::VectorXd dx = mK * innov;
  Eigen::MatrixXd mSS = mH * state.mP * mH.transpose() + mRm_ * 10.0 * 10.0;

  UpdateState(state, nextState, dx, mH, mK);
  return true;
}

Eigen::VectorXd RtkMeasure::MeasurementEquation(const State &state) const
{
  Eigen::VectorXd tempMeasure(3);
  tempMeasure = state.p;
  return tempMeasure;
}

void RtkMeasure::ObserveRTK(const State &state, State &nextState, Eigen::VectorXd &innov,
                            Eigen::MatrixXd &mH)
{
  innov = Eigen::Vector3d::Zero();
  innov = posUTM_ - (state.p);
  mH = Eigen::MatrixXd::Zero(3, StateIndex::STATE_TOTAL);
  mH.block<3, 3>(0, StateIndex::P) = Eigen::Matrix3d::Identity();
  mRm_ = Eigen::MatrixXd::Zero(3, 3);
  mRm_.block<3, 3>(0, 0) = gnssPVar_ * Eigen::Matrix3d::Identity();
}

} // namespace Ego
} // namespace Localizaiton

